import argparse
import decimal
import re

from pyunitx._api import _EXTANT_UNITS, Compound, _base_scale, make_compound_unit

_EXTANT_ABBREVS = {unit.abbreviation: unit for name, unit in _EXTANT_UNITS.items()}


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--figures", default=6, type=int, help="The number of significant"
                                                                     " figures to output")
    parser.add_argument("quantity", help="The number of the source unit.")
    parser.add_argument("src", help="""\
    The unit to convert from. Must be expressed as a series of unit symbols with exponents, 
    multiplied together using '.'. '/' for division is not allowed, instead use a negative 
    exponent on the unit. As an example, you would give meters per second as 'm.s^-1'.""")
    parser.add_argument("dest", help="The unit to convert to, in the same format.")
    return parser.parse_args()


def parse_unit(spec):
    pattern = re.compile(r"\.?([\w ]+)(\^-?\d+)?")
    pairs = []
    for match in re.finditer(pattern, spec):
        name = match.group(1)
        if name not in _EXTANT_ABBREVS:
            raise KeyError(f"{name} is not a recognized unit")
        if match.group(2):
            power = int(match.group(2)[1:])
        else:
            power = 1
        pairs.append((_EXTANT_ABBREVS[name], power))
    return Compound(tuple(pairs))


def main():
    args = parse_args()
    num = decimal.Decimal(args.quantity)
    sourcecomp = parse_unit(args.src)
    destcomp = parse_unit(args.dest)
    source = make_compound_unit(
        exponents=sourcecomp.to_pairs(),
        scale=_base_scale(sourcecomp)
    )
    dest = make_compound_unit(
        exponents=destcomp.to_pairs(),
        scale=_base_scale(destcomp)
    )
    if not source(num).is_dimension(dest.dimension):
        raise TypeError("Not the same dimension, cannot be converted")
    print(dest(num * source.scale / dest.scale).sig_figs(args.figures))


if __name__ == '__main__':
    main()



from pyunitx._api import make_dimension, make_unit, si_unit

Mass = make_dimension("Mass")

kilograms = make_unit(
    name="kilograms",
    dimension=Mass,
    scale=1,
    abbrev="kg",
    doc="""\
    Kilograms are the SI base unit of mass, despite having a magnitude prefix.
    """
)
grams = make_unit(
    name="grams",
    dimension=Mass,
    scale="0.001",
    abbrev="g",
    doc="""A gram is roughly defined as the mass of a milliliter of water."""
)

generated = si_unit(base_unit=grams, skip=["kilo"])
globals().update(generated)

av_pound = make_unit(
    name="avoirdupois_pounds_mass",
    dimension=Mass,
    scale="0.45359237",
    abbrev="lbm_A",
    doc="""Avoirdupois pounds are one of the definitions of mass in the |ucs|."""
)
troy_pound = make_unit(
    name="troy_pounds_mass",
    dimension=Mass,
    scale="0.3732417",
    abbrev="lbm_T",
    doc="""Troy pounds are a little smaller."""
)
slug = make_unit(
    name="slug",
    dimension=Mass,
    scale="14.59390",
    abbrev="slug",
    doc="""\
    A slug is the amount of mass that would be accelerated at 1 
    foot/second/second by a force of one pound. Using slugs you can form a 
    consistent set of measurements whereas the pound-mass cannot.
    """
)

__all__ = [
              "Mass",
              "kilograms",
              "av_pound",
              "troy_pound",
              "slug",
              "grams"
          ] + list(generated.keys())



"""Defines common units of time.

For better handling of dates and times, use :mod:`datetime` from the standard
library. This module is suitable for manipulation of time intervals and rates
that show up in other calculations but knows absolutely nothing about leap
seconds or even dates.
"""

from decimal import Decimal

from pyunitx._api import make_dimension, make_unit, si_unit

Time = make_dimension("Time")
seconds = make_unit(
    name="seconds",
    dimension=Time,
    scale=1,
    abbrev="s",
    doc="""\
    The second is the SI base unit of time. As there are convenient larger 
    units for time already, the positive SI prefixes (k, M, G) aren't used very 
    much but the negative ones (m, μ, n) do see common use.
    """
)

generated = si_unit(base_unit=seconds)
globals().update(generated)

minutes = make_unit(
    name="minutes",
    dimension=Time,
    scale=60,
    abbrev="min",
    doc="""One minute is sixty seconds."""
)
hours = make_unit(
    name="hours",
    dimension=Time,
    scale=60 * 60,
    abbrev="hr",
    doc="""One hour is sixty minutes."""
)
days = make_unit(
    name="days",
    dimension=Time,
    scale=60 * 60 * 24,
    abbrev="day",
    doc="""\
    A solar day is exactly 24 hours. The solar day is loosely defined as the 
    period between zeniths of the sun. 
    """
)
sidereal_days = make_unit(
    name="sidereal_days",
    dimension=Time,
    scale=60 * 60 * 23 + 60 * 56 + Decimal("4.091"),
    abbrev="day_sd",
    doc="""\
    While the earth is spinning, it is also orbiting the sun. The movement along
    the orbit over the course of the day means that the sun has moved backwards
    a bit so the earth has to do some extra rotation before the sun is in the 
    same place again. If you instead hovered above one of the poles with a fixed
    orientation, and tracked the time it took for the earth to rotate back into
    the same position, you would get the length of the sidereal day. This is a 
    little shorter than a solar day.
    """
)
julian_years = make_unit(
    name="julian_years",
    dimension=Time,
    scale=days.scale * Decimal("365.25"),
    abbrev="jyr",
    doc="""\
    It's nice to have a definition of a year that doesn't vary with things like
    leap years, and the Julian year is that.
    """
)
years = make_unit(
    name="years",
    abbrev="yr",
    dimension=Time,
    scale="3.155693e7",
    doc="""\
    This is the most common, solar definition of the year—the time it takes 
    for the earth to complete one orbit around the sun. Somewhat more precisely
    it is the interval between subsequent vernal equinoxes (or summer solstices,
    they're all identical).
    """
)

__all__ = [
              "Time",
              "seconds", "minutes", "hours",
              "days", "sidereal_days",
              "julian_years",
          ] + list(generated.keys())



from pyunitx._api import make_compound_unit, make_compound_dimension, si_unit
from pyunitx.current import Current, amperes
from pyunitx.voltage import Potential, volts


Resistance = make_compound_dimension(name="Resistance", exponents={Potential: 1, Current: -1})

ohms = make_compound_unit(
    name="ohms",
    abbrev="Ω",
    exponents={volts: 1, amperes: -1},
    scale=1,
    doc="""\
    A current of one ampere traveling through one ohm of resistance drops its
    potential by one volt.
    """
)

generated = si_unit(base_unit=ohms)

__all__ = [
              "Resistance",
              "ohms"
          ] + list(generated.keys())



from decimal import Decimal

from pyunitx._api import make_compound_unit
from pyunitx.energy import joules
from pyunitx.length import meters
from pyunitx.mass import kilograms
from pyunitx.mole import moles
from pyunitx.power import watts
from pyunitx.pressure import pascals
from pyunitx.temperature import kelvin
from pyunitx.time import seconds

__all__ = [
    "c", "speed_of_light",
    "G", "gravitational_constant",
    "N_A", "avogadro",
    "σ", "stefan_boltzmann",
    "R", "gas_constant",
    "M_air", "air_molar_mass",
    "g", "earth_gravity",
    "atm", "standard_atmosphere",
    "R_E", "earth_radius",
    "M_E", "earth_mass",
]


# Most units from
# https://physics.nist.gov/cuu/Constants/Table/allascii.txt

def _construct_gravitational_constant():
    g_unit = make_compound_unit(scale=1, exponents={meters: 3, kilograms: -1, seconds: -2})
    return g_unit("6.67430e-11")


def _construct_stefan_boltzmann_constant():
    unit = make_compound_unit(scale=1, exponents={watts: 1, meters: -2, kelvin: -4})
    return unit("5.670374419e-8")


def _construct_standard_gravity():
    unit = make_compound_unit(scale=1, exponents={meters: 1, seconds: -2})
    return unit("9.80665")


def _construct_gas_constant():
    unit = make_compound_unit(scale=1, exponents={joules: 1, moles: -1, kelvin: -1})
    return unit("8.31446261815324")


def _construct_air_molar_mass():
    unit = make_compound_unit(scale=1, exponents={kilograms: 1, moles: -1})
    # https://www.engineeringtoolbox.com/molecular-mass-air-d_679.html
    return unit(".0289647")


#: The speed of light in a vacuum *c* is an absolute constant of the universe.
#: It now forms the foundation for the definition of the meter.
#: It is given here in :math:`\frac{m}{s}`.
c = meters(299_792_458) / seconds(1)
speed_of_light = c

#: Newton's gravitational constant *G* shows up in the classical equation
#: :math:`F = \frac{GMm}{r^2}`.
#: It is given here in :math:`\frac{m^3}{kg s^2}`.
G = _construct_gravitational_constant()
gravitational_constant = G

#: Avogadro's number is the number of molecules making up one mole,
#: approximately :math:`6.022\times 10^{23}`
N_A = Decimal("6.02214076e23")
avogadro = N_A

#: The Stefan-Boltzmann constant relates the temperature of an object to how
#: much electromagnetic radiation it emits. It appears in the equation
#: :math:`j^* = \sigma T^4`.
#: It is given in units of :math:`\frac{W}{m^2 K^4}`.
σ = _construct_stefan_boltzmann_constant()
stefan_boltzmann = σ
sigma = σ

#: One standard atmosphere is approximately the average sea-level atmospheric
#: pressure.
atm = pascals(101325)
standard_atmosphere = atm

#: The ideal gas constant relates the pressure, volume, quantity, and
#: temperature of a gas to each other.
#: This equation is :math:`PV = nRT`.
#: It is given in :math:`\frac{J}{mol K}`.
R = _construct_gas_constant()
gas_constant = R

#: The average molar mass of dry air means the mass of one mole of air.
#: Factoring in molar mass means you can transform the ideal gas law into
#: :math:`PV = \frac{m}{M_{air}} R T` if you have a mass instead of molarity
#: which is the common case.
#: It is in units of :math:`\frac{kg}{mol}`
M_air = _construct_air_molar_mass()
air_molar_mass = M_air

#: Standard earth's gravity is an average of measurements taken around the
#: world. If you live at a high elevation, you may measure a lower acceleration
#: because you are further from the center of the earth.
#: It is in units of :math:`\frac{m}{s^2}`
g = _construct_standard_gravity()
earth_gravity = g

#: The earth's equatorial radius, as measured for the WGS 84 ellipsoid.
#: Its value is retrieved from
#: `here <https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html>`_.
#: It is given in :math:`m`.
R_E = meters(6378137)
earth_radius = R_E

#: The earth's total mass. Its value is retrieved from
#: `here <https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html>`_.
#: It is given in :math:`kg`.
M_E = kilograms("5.9722e24")
earth_mass = M_E



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.current import Current, amperes
from pyunitx.time import Time, seconds

Charge = make_compound_dimension(name="Charge", exponents={Current: 1, Time: 1})

coulombs = make_compound_unit(
    name="coulombs",
    abbrev="C",
    exponents={amperes: 1, seconds: 1},
    scale=1,
    doc="""\
    A coulomb is formally defined as the amount of charge that travels past a
    point in a 1-Ampere current over the course of one second. Current is the
    defined fundamental quantity in the SI system, and is certainly easier to
    measure directly, but charge cannot be decomposed.
    """
)

generated = si_unit(base_unit=coulombs)
globals().update(generated)

fundamental_charge = make_compound_unit(
    name="fundamental_charge",
    abbrev="e",
    exponents={amperes: 1, seconds: 1},
    scale="1.602176634e-19",
    doc="""\
    This is the charge of a single proton or electron.
    """
)

__all__ = [
              "Charge",
              "coulombs",
              "fundamental_charge",
          ] + list(generated.keys())



from pyunitx._api import make_unit, make_dimension

__all__ = [
    "Data",
    "bits",
    "bytes",
    "nybbles",
    "kibibytes",
    "mebibytes",
    "gibibytes",
    "tebibytes",
    "exbibytes",
]

Data = make_dimension("Data")

bits = make_unit(
    name="bits",
    abbrev="b",
    dimension=Data,
    scale=1,
    doc="""\
    A bit is the smallest possible unit of data, representing the state of a 
    system that has exactly two possible states.
    """
)
bytes = make_unit(
    name="bytes",
    abbrev="B",
    dimension=Data,
    scale=8,
    doc="""\
    The modern definition of a byte is 8 bits. Originally, this just meant "the
    smallest addressable unit of memory" and could vary between different 
    computers.
    """
)
nybbles = make_unit(
    name="nybbles",
    abbrev="N",
    dimension=Data,
    scale=4,
    doc="""\
    One nybble is half a byte, 4 bits. It is generally not used for bulk data,
    only really for low-level work where every bit counts. Some CPU 
    architectures allow nybble-level manipulations.
    """
)
kibibytes = make_unit(
    name="kibibytes",
    abbrev="KiB",
    dimension=Data,
    scale=8 * 1024,
    doc="""\
    One KiB is 1024 bytes.
    
    Use of the SI prefixes ``kilo``, etc. are discouraged with data as the most
    natural set of scales are based on powers of two. Thus prefixes are selected
    using 2\\ :sup:`10` = 1024, which is close to 1000.
    """
)
mebibytes = make_unit(
    name="mebibytes",
    abbrev="MiB",
    dimension=Data,
    scale=8 * 1024 ** 2,
    doc="""\
    One MiB is 1024\\ :sup:`2` bytes.
    
    Use of the SI prefixes ``kilo``, etc. are discouraged with data as the most
    natural set of scales are based on powers of two. Thus prefixes are selected
    using 2\\ :sup:`10` = 1024, which is close to 1000.
    """
)
gibibytes = make_unit(
    name="gibibytes",
    abbrev="GiB",
    dimension=Data,
    scale=8 * 1024 ** 3,
    doc="""\
    One GiB is 1024\\ :sup:`3` bytes.
    
    Use of the SI prefixes ``kilo``, etc. are discouraged with data as the most
    natural set of scales are based on powers of two. Thus prefixes are selected
    using 2\\ :sup:`10` = 1024, which is close to 1000.
    """
)
tebibytes = make_unit(
    name="tebibytes",
    abbrev="TiB",
    dimension=Data,
    scale=8 * 1024 ** 4,
    doc="""\
    One TiB is 1024\\ :sup:`4` bytes.
    
    Use of the SI prefixes ``kilo``, etc. are discouraged with data as the most
    natural set of scales are based on powers of two. Thus prefixes are selected
    using 2\\ :sup:`10`, which is close to 1000.
    """
)
exbibytes = make_unit(
    name="exbibytes",
    abbrev="EiB",
    dimension=Data,
    scale=8 * 1024 ** 5,
    doc="""\
    One EiB is 1024\\ :sup:`5` bytes.
    
    Use of the SI prefixes ``kilo``, etc. are discouraged with data as the most
    natural set of scales are based on powers of two. Thus prefixes are selected
    using 2\\ :sup:`10` = 1024, which is close to 1000.
    """
)



from pyunitx._api import make_dimension, make_unit, si_unit

Length = make_dimension("Length")

meters = make_unit(
    name="meters",
    dimension=Length,
    scale=1,
    abbrev="m",
    doc="""\
    Meters are the base SI unit of length. It is currently defined in terms of
    the speed of light.
    """
)

generated = si_unit(base_unit=meters)
globals().update(generated)

feet = make_unit(
    name="feet",
    dimension=Length,
    scale="0.3048",
    abbrev="ft",
    doc="""Feet are the base unit of length in the |ucs|."""
)
miles = make_unit(
    name="miles",
    dimension=Length,
    scale="1609.344",
    abbrev="mi",
    doc="""Miles are the long distance measure in the |ucs|."""
)
yards = make_unit(
    name="yards",
    dimension=Length,
    scale="0.9144",
    abbrev="yd",
    doc="""One yard is equal to three feet."""
)
inches = make_unit(
    name="inches",
    dimension=Length,
    scale="0.0254",
    abbrev="in",
    doc="""Twelve inches make up one foot."""
)
astronomical_unit = au = make_unit(
    name="au",
    dimension=Length,
    scale=149_597_870_700,
    abbrev="au",
    doc="""\
    One astronomical unit, or AU, is the average distance between Earth and the
    Sun. It's used in astronomical measurements, and is part of the definition
    of the :class:`parsec <pyunitx.derived.parsec>`.
    """
)

__all__ = [
              "Length",
              "meters",
              "feet",
              "miles",
              "yards",
              "inches",
              "astronomical_unit",
              "au",
          ] + list(generated.keys())



from pyunitx._api import make_compound_unit, make_compound_dimension, si_unit
from pyunitx.charge import Charge, coulombs
from pyunitx.energy import Energy, joules

Potential = make_compound_dimension(name="Potential", exponents={Energy: 1, Charge: -1})

volts = make_compound_unit(
    name="volts",
    abbrev="V",
    scale=1,
    exponents={joules: 1, coulombs: -1},
    doc="""\
    Volts are the only real unit of electrical potential energy.
    """
)

generated = si_unit(base_unit=volts)
globals().update(generated)

__all__ = [
              "Potential",
              "volts",
          ] + list(generated.keys())



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.time import Time, seconds, minutes

Frequency = make_compound_dimension({Time: -1}, "Frequency")
hertz = make_compound_unit(
    name="hertz",
    scale=1,
    exponents={seconds: -1},
    abbrev="Hz",
    doc="""\
    The hertz represents an occurrence once per second. Most commonly,
    this is used to measure sounds, where the occurrence is a soundwave peak 
    passing. 
    """
)

generated = si_unit(base_unit=hertz)
globals().update(generated)

rpm = make_compound_unit(
    name="rpm",
    scale=1 / 60,
    exponents={minutes: -1},
    abbrev="rpm",
    doc="""\
    Revolutions per minute measure the action of a piston crank or of a wheel.
    """
)

__all__ = [
              "Frequency",
              "hertz",
              "rpm",
          ] + list(generated.keys())



from pyunitx._api import make_unit
from pyunitx.angle import arcsecond
from pyunitx.constants import c
from pyunitx.length import Length, au
from pyunitx.time import julian_years

__all__ = ["parsec", "lightyear"]

parsec_calc = au(1) / arcsecond(1).to_radians()
parsec = make_unit(
    name="parsec",
    dimension=Length,
    scale=parsec_calc.to_meters().value,
    abbrev="pc",
    doc="""\
    One parsec is an astronomical distance defined by observation of the stars.
    As we orbit the sun, nearer stars move with respect to the extremely distant
    ones, an effect called parallax. Since we know the distance from the earth
    to the sun, we can calculate the distance the distance to those stars
    with a little trigonometry.
    
    If a star moves across an angle of one 
    :class:`arcsecond <pyunitx.angle.arcsecond>` as we observe it at either end of
    a distance of 1 AU, that means it is one parsec away. 
    """
)
lightyear = make_unit(
    name="lightyear",
    dimension=Length,
    scale=c.value * julian_years.scale,
    abbrev="ly",
    doc="""\
    A lightyear is the distance light travels in one year.
    As the real duration of a year changes over time (we're not going to say a
    lightyear is a different distance in a leap year) it uses a constant value
    of a Julian year, or 365.25 days.
    """
)



import itertools
import math
import re
import textwrap
import warnings
from decimal import Decimal
from numbers import Number
from typing import Union, Tuple, Sequence, Type, Dict, Iterator, Optional

import sigfig

from pyunitx._exceptions import OperationError, ImplicitConversionError

__all__ = [
    "make_dimension",
    "make_unit",
    "make_compound_dimension",
    "make_compound_unit",
    "si_unit",
    "SIUNITX_NEW",
    "SIUNITX_OLD",
]
UnitOperand = Union['UnitInterface', Number]
Scale = Union[Decimal, float, str, Tuple[int, Sequence[int], int]]
Unitlike = Union[Type['UnitInterface'], 'DimensionBase']
Pair = Tuple[Unitlike, int]
Pairs = Tuple[Pair, ...]
Exponents = Union[Pairs, Dict[Unitlike, int]]

SIUNITX_NEW = 3
SIUNITX_OLD = 2
_SI_PREFIXES = [
    ["yotta", "Y", "1e24"],
    ["zetta", "Z", "1e21"],
    ["exa", "E", "1e18"],
    ["peta", "P", "1e15"],
    ["tera", "T", "1e12"],
    ["giga", "G", "1e9"],
    ["mega", "M", "1e6"],
    ["kilo", "k", "1e3"],
    ["hecto", "h", "1e2"],
    ["deka", "da", "1e1"],
    ["deci", "d", "1e-1"],
    ["centi", "c", "1e-2"],
    ["milli", "m", "1e-3"],
    ["micro", "μ", "1e-6"],
    ["nano", "n", "1e-9"],
    ["pico", "p", "1e-12"],
    ["femto", "f", "1e-15"],
    ["atto", "a", "1e-18"],
    ["zepto", "z", "1e-21"],
    ["yocto", "y", "1e-24"],
]

_EXTANT_UNITS = {}


def make_dimension(name: str) -> 'DimensionBase':
    """Dimensions are the basic measurable quantities about the world, like length and time.

    For complex dimensions, see :func:`make_compound_dimension`.

    :param name: The name of this dimension.
    :return: An object that captures information about the dimension and the
        units that implement it.
    """
    dimension = make_compound_dimension((), name)
    return dimension


def make_unit(*, name: str, dimension: 'DimensionBase', scale: Scale, abbrev: str, doc="") \
        -> Type['UnitInterface']:
    """A unit is a particular convention for measuring a dimension.

    The resulting class represents the abstract concept of the unit, say meters.
    Instances of the class associate that unit with a number, defining in this
    example a particular length like 8 meters. Instances of a unit class are
    called "measurements" within this documentation. Instances are immutable;
    any interactions with the public API will not change the object, anything
    that appears like it (for instance rounding with
    :meth:`sig_figs <pyunitx.length.meters.sig_figs>`) instead constructs and
    returns a copy.

    :param name: The name of the new unit, like "meters". Should be in its
        plural form.
    :param dimension: The dimension this unit measures.
    :param scale: How many of the base unit it would take to get one of the
        current unit. For instance, one kilometer has a scale of 1000 when the
        base unit is meters.
    :param abbrev: The canonical abbreviation for this unit, like "lb" for
        pounds or "g" for grams. Officially called the unit's symbol. Used for
        output.
    :param doc: Documentation for this unit, optional.
    :return: A class representing the unit. Instances of this class are
        measurements using the unit.
    """

    def new(cls, value: Scale):
        """Create a new measurement using this unit.

        Instances are flyweights; two invocations of ``meters(1)`` will return
        the same object.

        :param value: The numerical value of the measurement, in any form that
            :external:py:class:`decimal.Decimal` can accept.
        :return: The newly created measurement, or the cached version.
        """
        v = Decimal(value)
        if v not in cls.instances:
            cls.instances[v] = super(type, cls).__new__(cls)
        instance = cls.instances[v]
        instance.value = v
        return instance

    def add(self, other: UnitInterface) -> UnitInterface:
        """Add two measurements of the same unit together.

        :param other: Another measurement, with the same units.
        :raises TypeError: If measurements are not the same unit.
        :return: A measurement with the values of the two added.
        """
        if type(other) != type(self):
            raise OperationError("add", type(self), type(other))
        return type(self)(self.value + other.value)

    def subtract(self, other: UnitInterface) -> UnitInterface:
        """Subtract a measurement of the same unit from this.

        :param other: Another measurement, with the same units.
        :raises TypeError: If measurements are not the same unit.
        :return: A measurement with the values of the two subtracted.
        """
        if type(other) != type(self):
            raise OperationError("subtract", type(self), type(other))
        return type(self)(self.value - other.value)

    def equal(self, other: UnitInterface) -> bool:
        """Check if this measurement is the same value and unit of another.

        :param other: Another measurement, with the same units.
        :return: Whether these two measurements are identical.
        """
        if self is other:
            return True
        if type(other) != type(self):
            return False
        return self.value == other.value

    def equivalent(self, other: UnitInterface, figs=5) -> bool:
        """Check if this measurement represents the same quantity as another,
        within a certain precision.

        This effectively converts both units to the base unit to get them on
        equal footing before rounding and checking equality.

        :param other: The other measurement.
        :param figs: How many significant figures to round to before comparison.
        :raises TypeError: If the two units do not represent the same dimension
            and therefore cannot be compared.
        :return: Whether the two measurements are approximately equal.
        """
        if not other.is_dimension(self.dimension):
            raise ImplicitConversionError(type(other), type(self))
        with warnings.catch_warnings():
            # sigfig.round raises warnings if the quantity being rounded has fewer significant
            # figures than are given, but for the purposes of this comparison, I don't care
            # so ignore them all
            warnings.simplefilter("ignore")
            a = sigfig.round(self.value * self.scale, sigfigs=figs)
            b = sigfig.round(other.value * other.scale, sigfigs=figs)
        return a == b

    def multiply(self, other: Union[UnitInterface, int, float, Decimal]) -> UnitInterface:
        """Multiply two measurements.

        If the two quantities completely cancel (like a frequency times a
        duration), the result will be returned as a plain number.

        Otherwise, produce the correct unit by combining the units of the two
        multiplicands. If the result has two base units that measure the same
        dimension but are not the same (say you're trying to multiply ``ft/s`` by
        ``m^2``, ``ft*m^2`` would be in the result) a TypeError will be thrown.
        Instead you should convert one of the measurements to be compatible with
        the other before multiplying. In the example, that could be converting the
        ``ft/s`` to ``m/s``.

        :param other: Another measurement, with any units including none.
        :raises TypeError: If base units are incompatible.
        :return: The result of the calculation, with the compound units.
        """
        if isinstance(other, (int, float, Decimal)):
            return type(self)(self.value * Decimal(other))
        result_value = self.value * other.value
        unit_composition = (self.composition * other.composition).units
        if len(unit_composition) == 0:
            return result_value
        result_unit = make_compound_unit(
            scale=self.scale * other.scale,
            exponents=unit_composition
        )
        return result_unit(result_value)

    def divide(self, other: Union[UnitInterface, int, float, Decimal]) -> UnitInterface:
        """Divide two measurements.

        If the two quantities completely cancel (like the derivation of radians
        dividing length by length), the result will be returned as a plain
        number.

        Otherwise, produce the correct unit by combining the units of the two
        divisors. If the result has two base units that measure the same
        dimension but are not the same (say you're trying to divide ``ft/s`` by
        ``m^2``, ``ft/m^2`` would be in the result) a TypeError will be thrown.
        Instead you should convert one of the measurements to be compatible with
        the other before multiplying. In the example, that could be converting the
        ``ft/s`` to ``m/s``.

        :param other: Another measurement, with any units including none.
        :raises TypeError: If base units are incompatible.
        :return: The result of the division, with correct units.
        """
        if isinstance(other, (int, float, Decimal)):
            return type(self)(self.value / Decimal(other))
        result_units = self.composition / other.composition
        result_value = self.value / other.value
        if len(result_units) == 0:
            return result_value
        result_unit = make_compound_unit(
            scale=self.scale / other.scale,
            exponents=result_units.to_pairs()
        )
        return result_unit(result_value)

    # rdivide can be much simpler because dividing two units will be handled by main divide
    def rdivide(self, other: Union[int, float, Decimal]) -> UnitInterface:
        result_unit = type(self ** -1)
        return result_unit(Decimal(other) / self.value)

    def exponent(self, other: Scale) -> UnitInterface:
        """Raise this measurement to a power such that the result has no
        non-integer exponents.

        For instance, ``meters(4) ** 2 == meters_squared(16)``. You can also
        compute the magnitude of a vector with
        ``vec = [meters(4), meters(3)]; mag = (vec[0] ** 2 + vec[1] ** 2) ** (1/2)``.

        If the resulting calculation has any units with non-integer powers,
        ``ValueError`` is raised.

        :param other: Any integer, the power to raise this measurement to.
        :raises ValueError: If the resulting unit has fractional exponents.
        :return: A measurement with the value and units raised to the power.
        """
        v = Decimal(other)
        result_composition = self.composition ** v
        result_value = self.value ** v
        result_unit = make_compound_unit(
            scale=self.scale ** v,
            exponents=result_composition.to_pairs()
        )
        return result_unit(result_value)

    def less(self, other: UnitInterface):
        """Check if this measurement is less than another.

        :param other: The measurement to compare to.
        :raises TypeError: If the two arguments are different units and can
            therefore not be compared.
        :return: Whether the other measurement is less than this one.
        """
        if type(other) != type(self):
            raise TypeError(f"{self.abbreviation} and {other.abbreviation} cannot be compared")
        return self.value * self.scale < other.value * other.scale

    def lesseq(self, other: UnitInterface):
        """Check if this measurement is less than or equal to another.

        :param other: The measurement to compare to.
        :raises TypeError: If the two arguments are different units and can
            therefore not be compared.
        :return: Whether the other measurement is less than this one.
        """
        return self < other or self == other

    def absolute(self):
        """Get the absolute value of this measurement.

        :return: A copy of this measurement as an absolute value.
        """
        return type(self)(abs(self.value))

    def neg(self):
        """Get the negative of this measurement.

        :return: A copy of this measurement with sign flipped.
        """
        return type(self)(-self.value)

    def pos(self):
        """No-op, return self.

        This is unlike any other math operation since it doesn't make a copy.
        """
        return self

    def is_dimension(self, dim: DimensionBase):
        """Check if this unit is of the given dimension.

        Most useful for checking whether two measurements can be reasonably
        compared with ``.equivalent_to`` or converted into the other.

        :param dim: The dimension to check against.
        :return: Whether this unit is of the given dimension.
        """
        try:
            return dim.composition == self.dimension.composition
        except AttributeError:
            return False

    def getattribute(self, key: str):
        """Forward conversion requests to the dimension.

        The dimension, being the one thing all compatible units have in common,
        is the logical place to store all conversion functions between different
        units.

        Any ``x.to_*`` method calls are therefore passed to the dimension.

        :param key: The method name.
        :return: A pseudo-bound method of the conversion function.
        :raise AttributeError: If the requested method doesn't look like a
            conversion function, or if there isn't a conversion function by that
            name.
        """
        if key.startswith("to_"):
            if not hasattr(self.dimension, key):
                composition = Compound.from_string(key[3:])
                base_scale = _base_scale(composition)
                unit = make_compound_unit(
                    name=key[3:],
                    scale=base_scale,
                    exponents=composition.to_pairs(),
                )
                key = "to_" + unit.__name__
                # It is automatically registered on the dimension on creation,
                # no need to explicitly use it
            return lambda: getattr(self.dimension, key)(self)
        raise AttributeError()

    def tostring(self):
        return f"{self.value} {self.abbreviation}"

    def sig_figs(self, figs=3) -> UnitInterface:
        """Produce a version of this measurement rounded to significant figures.

        :param figs: How many significant figures to round to.
        :return: A new measurement with a rounded value.
        """
        rounded = sigfig.round(self.value, sigfigs=figs)
        return type(self)(rounded)

    def to_latex(self, siunitx_major_version=SIUNITX_NEW):
        r"""Output this unit value as it would be used in LaTeX with siunitx.

        Maybe script your calculations, then write the output to a .tex file
        and include it in your main document.

        :param siunitx_major_version: siunitx changed the name of its macro
            that outputs a number with a unit in version 3.0. Prior, it was
            ``\SI{number}{units}``, and after it was ``\qty{number}{units}``.
            Defaults to the most recent version. If you're using an older one,
            for instance if your TeXLive distribution is a few years old, pass
            in :data:`pyunitx.SIUNITX_OLD`, an alias for 2. I doubt anyone is
            using version 1.
        :return: A string formatted for direct inclusion in LaTeX.
        """
        if siunitx_major_version >= 3:
            fmt = r"\qty{{{value}}}{{{unit}}}"
        else:
            fmt = r"\SI{{{value}}}{{{unit}}}"
        if re.match(r"^\w+$", self.abbreviation):
            # This unit has an elementary name, like J for joules
            return fmt.format(value=self.value, unit=self.abbreviation)
        units = []
        for u, e in self.composition.to_pairs():
            if e == 1:
                units.append(u.abbreviation)
            else:
                units.append(f"{u.abbreviation}^{{{e}}}")
        return fmt.format(value=self.value, unit=".".join(units))

    def closest_si_prefix(self):
        """Convert this value to the most natural SI prefix.

        For instance, a value of 12100 meters would be changed into 12.1
        kilometers.
        This only works on units that have a specific name. Elementary units
        like meters of course work, as do composite units like joules. However,
        if a unit can only be expressed by some combination of named units
        (say N*s) then a TypeError will be raised.

        Currently ignores 10^-2 through 10^2; that is, centi through hecto.
        These are not used nearly as often, with the exception of the
        centimeter.

        :raises TypeError: If this unit is not an SI unit.
        :raises ValueError: If the value of this unit is too big or too small
            to be directly expressed with an SI prefix. Instead you should
            probably express your value as scientific notation using the base
            unit of the dimension.
        :return: This value expressed in the unit that requires no extra
            scientific notation.
        """
        magnitude = math.log10(self.scale)
        if magnitude != int(magnitude):
            raise TypeError("This isn't an SI unit so prefixes can't be applied")
        pattern = re.compile(
            f"[{''.join(p for _, p, __ in _SI_PREFIXES)}]?" +
            "(m|g|s|A|K|cd|mol|J|Hz|C|Pa|V|Ω)"
        )
        if not pattern.match(self.abbreviation):
            raise TypeError("This isn't a base SI unit so prefixes can't be applied")
        order = (self.value * self.scale).adjusted()
        if order > 26 or order < -24:
            raise ValueError("SI prefixes only cover 48 orders of magnitude")
        mag = (order // 3) * 3
        family = [u for u in _EXTANT_UNITS.values()
                  if u.dimension == self.dimension and u.scale == Decimal(f"1e{mag}")]
        if len(family) != 1:
            # kelvin/celsius are a special case, being equal in scale
            k = [u for u in family if u.__name__ == "kelvin"]
            family = [k[0]]
        converter_name = f"to_{family[0].__name__}"
        return getattr(self, converter_name)()

    # @formatter:off
    # noinspection PyTypeChecker
    unit: Type[UnitInterface] = type(name, (object,), {
        # Initializers
        "__new__": new,
        # Math
        "__add__": add,
        "__radd__": add,
        "__sub__": subtract,
        "__rsub__": subtract,
        "__abs__": absolute,
        "__pos__": pos,
        "__neg__": neg,
        "__mul__": multiply,
        "__rmul__": multiply,
        "__truediv__": divide,
        "__rtruediv__": rdivide,
        "__pow__": exponent,
        # Get special conversion functions
        "__getattr__": getattribute,
        # Comparisons; python can figure out the rest from this minimal set
        "__eq__": equal,
        "__lt__": less,
        "__le__": lesseq,
        # Stringifiers
        "__str__": tostring,
        "__repr__": tostring,
        # Special class variables
        "__name__": name,
        "__doc__": textwrap.dedent(doc),
        # Class variables
        "abbreviation": abbrev,
        "scale": Decimal(scale),
        "instances": {},
        "dimension": dimension,
        # Methods
        "is_dimension": is_dimension,
        "equivalent_to": equivalent,
        "sig_figs": sig_figs,
        "to_latex": to_latex,
        "to_natural_si": closest_si_prefix,
    })
    # @formatter:on
    unit.composition = Compound(((unit, 1),))

    def converter(self):
        f"""Convert the current unit to {name}"""
        return unit(self.value * self.scale / unit.scale)

    setattr(dimension, "to_" + name.replace(" ", "_"), converter)
    _EXTANT_UNITS[unit.__name__] = unit
    return unit


def make_compound_dimension(exponents: Exponents, name: str = None) -> 'DimensionBase':
    """Create a dimension that's composed of basic dimensions.

    :param exponents: A sequence of tuples or dict that pair existing
        dimensions with the exponent they should be raised to.
    :param name: A name to give this dimension, like velocity for length/time.
        If there isn't a physically useful name for it, leave it unfilled and
        a name will automatically be created by using the names of its
        constituents.
    :return: An object that captures information about the dimension and the
        units that implement it.
    """
    if isinstance(exponents, dict):
        exponents = tuple(exponents.items())
    if name is None:
        name = str(Multiset(exponents))
    dimension = DimensionBase(name, exponents)

    return dimension


def make_compound_unit(*, scale: Scale, exponents: Exponents, name: str = None, abbrev=None,
                       doc="") -> Type['UnitInterface']:
    """Create or retrieve a unit composed of other units.

    If an identical unit has already been created (defined by same exponents
    and scale) then return that pre-constructed unit. This makes sure that 
    calculations that result in a named unit (such as ohms) will be correctly
    reported as such rather than breaking down into their base components.

    :param scale: How many of the base unit one of this unit is equivalent to.
        Note that the scale factors of the constituent units are not taken into
        account. For instance, even though the conversion factor from meters per
        second to miles per hour could be determined strictly from the scale
        factors already defined for miles -> meters and hours -> seconds, doing
        that calculation is up to the constructor of the unit. I think this
        obeys the principle of least surprise, but I might be wrong.
    :param exponents: A sequence of tuples or dict that pair existing units
        with the exponent they should be raised to.
    :param name: A name to give this unit, like joules. If no particular
        physical name exists, leave it unfilled and a name will be automatically
        constructed using the names of its component units and their exponents.
    :param abbrev: A canonical abbreviation for this unit, like J for joules.
        Officially called the unit's symbol. If no particular one exists, leave
        it unfilled and it will be automatically be constructed using the
        symbols of its base units.
    :param doc: Documentation for this unit, to show in tools like Sphinx.
    :return: A class representing this unit. Instances of the class are
        measurements with a magnitude.
    """
    if isinstance(exponents, dict):
        exponents = tuple(exponents.items())
    composition = Compound(exponents)
    existing = _access_unit_cache(composition, scale)
    if existing:
        return existing
    if name is None:
        name = str(Multiset(exponents))
    dims_extracted = tuple((unit.dimension, exp) for unit, exp in composition.to_pairs())
    dims = _sort(_dedupe(dims_extracted))
    dimension = make_compound_dimension(dims)
    unit = make_unit(
        name=name,
        dimension=dimension,
        scale=scale,
        abbrev=abbrev if abbrev is not None else composition.make_abbreviation(),
        doc=doc
    )
    unit.composition = composition
    return unit


def si_unit(*, base_unit: Type['UnitInterface'], skip=()) \
        -> Dict[str, Type['UnitInterface']]:
    """Create the full range of SI prefixes on a unit.

    :param base_unit: The unit to which prefixes can be applied.
    :param skip: If you've already created one of the units (I did this with
        kilograms since prefixes apply to grams), list the prefix here so it
        doesn't get overwritten.
    :return: A dictionary between the name of the unit and the unit class.
    """
    generated = {}
    for prefix, short, scale in _SI_PREFIXES:
        if prefix in skip:
            continue
        new_scale = Decimal(scale) * base_unit.scale
        new_name = prefix + base_unit.__name__
        new_abbrev = short + base_unit.abbreviation
        if base_unit.composition == Compound(Multiset({base_unit: 1})):
            new_unit = make_unit(
                name=new_name,
                abbrev=new_abbrev,
                scale=new_scale,
                dimension=base_unit.dimension,
            )
        else:
            new_unit = make_compound_unit(
                name=new_name,
                abbrev=new_abbrev,
                scale=new_scale,
                exponents=base_unit.composition.to_pairs(),
            )
        generated[new_name] = new_unit
    return generated


def _exponent_name(unit: type, exponent: int) -> str:
    value_names = {
        1: "",
        2: "_squared",
        3: "_cubed",
        4: "_to_the_fourth",
        5: "_to_the_fifth",
        # that's the highest I've ever seen
    }
    prefix = 'per_' if exponent < 0 else ''
    body = unit.__name__.rstrip("s") if exponent < 0 else unit.__name__
    return f"{prefix}{body}{value_names[abs(exponent)]}"


def _decompose_all(exponents: Pairs) -> Pairs:
    accumulator = []
    for unitish, exponent in exponents:
        accumulator.extend(_decompose(unitish, exponent))
    return _sort(accumulator)


def _decompose(unit: Unitlike, factor: int) -> Pairs:
    if not hasattr(unit, "composition"):
        # must be a unit currently under construction and doesn't have composition yet
        return (unit, factor),
    if len(unit.composition) == 1:
        return tuple((k, e * factor) for k, e in unit.composition.to_pairs())
    accumulator = []
    for k, e in unit.composition.to_pairs():
        accumulator.extend(_decompose(k, e * factor))
    return tuple(accumulator)


def _dedupe(pairs: Iterator[Pair]) -> Dict[Unitlike, int]:
    accumulator = {}
    for unitish, num in pairs:
        if unitish in accumulator:
            accumulator[unitish] += num
        else:
            accumulator[unitish] = num
    return accumulator


def _sort(d: Union[Exponents, Iterator[Pair]]) -> Pairs:
    if isinstance(d, dict):
        pairs = tuple(d.items())
    else:
        pairs = d
    return tuple(sorted(pairs, key=lambda p: (-p[1], p[0].__name__)))


def _access_unit_cache(c: 'Compound', scale) -> Optional[Type['UnitInterface']]:
    remap = {(_sort(_dedupe(unit.composition.to_pairs())), unit.scale): unit
             for name, unit in _EXTANT_UNITS.items()}
    key = (_sort(_dedupe(c.to_pairs())), scale)
    if key in remap:
        return remap[key]
    return None


def _base_scale(composition: 'Compound') -> Decimal:
    base_scale = Decimal(1)
    for u, e in composition.to_pairs():
        base_scale *= u.scale ** e
    return base_scale


class DimensionBase:
    __INSTANCES = {}

    def __new__(cls, name: str, exponents: Pairs):
        exponents = tuple((k, v) for k, v in exponents if v != 0)
        if len(exponents) == 0:
            # Base dimensions are identified by name
            key = name
        elif len(exponents) == 1 and exponents[0][1] == 1:
            # If you end up with a base unit after a calculation
            decomposed = _decompose_all(exponents)
            if len(decomposed) == 1 and decomposed[0][1] == 1:
                key = name
            else:
                key = decomposed
        else:
            # Compound dimensions with the same base dimensions must refer to the same thing
            key = _decompose_all(exponents)
        if key not in cls.__INSTANCES:
            cls.__INSTANCES[key] = super(DimensionBase, cls).__new__(cls)
            this = cls.__INSTANCES[key]
            this.__name__ = name
            this.composition = Compound(exponents or ((this, 1),))
        return cls.__INSTANCES[key]

    def __hash__(self):
        return hash(self.__name__)

    def __repr__(self):
        return self.__name__


class Multiset:
    # just different enough from collections.Counter to be worth writing
    def __init__(self, pairs: Union[Pairs, 'Multiset', dict]):
        if isinstance(pairs, Multiset):
            self.store = pairs.store.copy()
        # Split these cases from each other just because it makes pycharm's
        # type inference work properly
        elif isinstance(pairs, dict):
            self.store = pairs.copy()
        else:
            self.store = _dedupe(pairs)
        self.store = {k: v for k, v in self.store.items() if v != 0}

    def __iter__(self) -> Iterator[Unitlike]:
        return iter(self.store.keys())

    def __getitem__(self, item):
        return self.store[item]

    def __eq__(self, other):
        if isinstance(other, Multiset):
            return self.store == other.store
        return False

    def __len__(self):
        return len(self.store)

    def __str__(self):
        names = [_exponent_name(k, v) for k, v in _sort(self.store)]
        return "_".join(names)

    def add(self, elem: Union[type, 'Multiset']):
        if isinstance(elem, type):
            elem = Multiset({elem: 1})
        return self.__merge(elem)

    def remove(self, elem: Union[type, 'Multiset']):
        if isinstance(elem, type):
            elem = Multiset({elem: -1})
        else:
            elem = Multiset({key: -value for key, value in elem.store.items()})
        return self.__merge(elem)

    def to_pairs(self) -> Pairs:
        return _sort((key, self[key]) for key in self)

    def __merge(self, other: 'Multiset') -> 'Multiset':
        copy = self.store.copy()
        for key in other:
            if key in copy:
                copy[key] += other[key]
            else:
                copy[key] = other[key]
            if copy[key] == 0:
                del copy[key]
        return Multiset(copy)


class Compound:
    def __init__(self, units: Union[Pairs, Multiset]):
        if isinstance(units, Multiset):
            self.units = Multiset(units)
        else:
            decomposed = _decompose_all(units)
            self.units = Multiset(decomposed)

    def __eq__(self, other):
        try:
            return self.units == other.units
        except AttributeError:
            return False

    def __len__(self):
        return len(self.units)

    def __mul__(self, other: Union[type, Multiset, 'Compound']) -> 'Compound':
        if isinstance(other, type):
            other = Multiset({other: 1})
        elif isinstance(other, Compound):
            other = other.units
        self.__verify_no_dimension_mismatch(other)
        return Compound(self.units.add(other))

    def __truediv__(self, other: Union[type, Multiset, 'Compound']) -> 'Compound':
        if isinstance(other, type):
            other = Multiset({other: 1})
        elif isinstance(other, Compound):
            other = other.units
        self.__verify_no_dimension_mismatch(other)
        return Compound(self.units.remove(other))

    def __pow__(self, power: Union[int, float, Decimal]):
        pairs = []
        for t, e in self.to_pairs():
            new_exponent = e * power
            if int(new_exponent) != new_exponent:
                raise ValueError(
                    f"The exponentiation has caused a non-integer exponent on {t}, "
                    f"which is not allowed."
                )
            pairs.append((t, int(new_exponent)))
        return Compound(tuple(pairs))

    def __verify_no_dimension_mismatch(self, extra: Multiset):
        existing_dimensions = {unit.dimension: unit for unit in self.units}
        for unit in extra:
            if (unit.dimension in existing_dimensions
                    and existing_dimensions[unit.dimension] is not unit):
                # the dimension is already represented in the current unit, but it isn't the
                # same unit
                raise ImplicitConversionError(unit, existing_dimensions[unit.dimension])

    def to_pairs(self):
        return self.units.to_pairs()

    def make_abbreviation(self) -> str:
        ordered = _sort(self.to_pairs())
        chunks = []
        for unit, exponent in ordered:
            if exponent == 1:
                chunks.append(unit.abbreviation)
            else:
                chunks.append(f"{unit.abbreviation}^{exponent}")
        return " ".join(chunks)

    @classmethod
    def from_string(cls, spec: str) -> 'Compound':
        name_values = {
            None: 1,
            "": 1,
            "_squared": 2,
            "_cubed": 3,
            "_to_the_fourth": 4,
            "_to_the_fifth": 5,
            # that's the highest I've ever seen
        }
        names = "|".join(
            itertools.chain(
                _EXTANT_UNITS,
                [n.rstrip("s") for n in _EXTANT_UNITS]
            )
        )
        pattern = re.compile(
            r"(per_)?(" + names + "|[a-z]+)(_squared|_cubed|_to_the_fourth|_to_the_fifth)?"
        )
        pairs = []
        for match in re.finditer(pattern, spec):
            name = match.group(2)
            if name not in _EXTANT_UNITS:
                name = name + "s"
                if name not in _EXTANT_UNITS:
                    raise KeyError(f"The requested unit {name} does not exist")
            unit = _EXTANT_UNITS[name]
            power = name_values[match.group(3)]
            if match.group(1):
                power *= -1
            pairs.append((unit, power))
        return cls(tuple(pairs))


class UnitInterface:
    """A dummy class for type checking purposes, defining the interface of a unit class."""
    composition: 'Compound'
    scale: 'Decimal'
    abbreviation: str
    __slots__ = ["value"]

    def __init__(self, value: Scale):
        ...

    def __add__(self, other: UnitOperand) -> 'UnitInterface':
        ...

    def __sub__(self, other: UnitOperand) -> 'UnitInterface':
        ...

    def __mul__(self, other: UnitOperand) -> 'UnitInterface':
        ...

    def __truediv__(self, other: UnitOperand) -> 'UnitInterface':
        ...

    def is_dimension(self, dim: DimensionBase) -> bool:
        ...

    def equivalent_to(self, quantity: 'UnitInterface', within=0) -> bool:
        ...

    def sig_figs(self, figures=3):
        ...

    def to_latex(self, siunitx_major_version=SIUNITX_NEW):
        ...

    def to_natural_si(self):
        ...



from pyunitx._api import si_unit, make_unit, make_dimension

Current = make_dimension("Current")

amperes = make_unit(
    name="amperes",
    abbrev="A",
    scale=1,
    dimension=Current,
    doc="""\
    An ampere is a rate of charge movement equal to one coulomb per second. It
    is one of the fundamental units in the SI system.
    """
)

generated = si_unit(base_unit=amperes)
globals().update(generated)

__all__ = [
              "Current",
              "amperes",
          ] + list(generated.keys())



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.length import Length, meters, feet
from pyunitx.mass import Mass, kilograms, slug
from pyunitx.time import Time, seconds

Force = make_compound_dimension(((Mass, 1), (Length, 1), (Time, -2)))

newtons = make_compound_unit(
    name="newtons",
    scale=1,
    exponents={kilograms: 1, meters: 1, seconds: -2},
    abbrev="N",
    doc="""The newton is the base SI unit of force."""
)

generated = si_unit(base_unit=newtons)
globals().update(generated)

kgf = kilograms_force = make_compound_unit(
    name="kilograms_force",
    scale="9.80665",
    exponents={kilograms: 1, meters: 1, seconds: -2},
    abbrev="kgf",
    doc="""One kgf is the weight of a kilogram in standard earth gravity."""
)
pounds = make_compound_unit(
    name="pounds",
    scale="4.4482216152605",
    exponents={slug: 1, feet: 1, seconds: -2},
    abbrev="lb",
    doc="""\
    Pounds are a more elementary unit in |ucs| than an actual mass.
    """
)
__all__ = ["Force", "newtons", "pounds", "kgf", "kilograms_force"] + list(generated.keys())



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.force import Force, pounds, newtons
from pyunitx.length import Length, meters, inches

Pressure = make_compound_dimension({Force: 1, Length: -2})

pascals = make_compound_unit(
    name="pascals",
    abbrev="Pa",
    scale=1,
    exponents={newtons: 1, meters: -2},
    doc="""The pascal is the base unit of pressure, but it is very small."""
)

generated = si_unit(base_unit=pascals)
globals().update(generated)

bars = make_compound_unit(
    scale="1e5",
    exponents={newtons: 1, meters: -2},
    name="bars",
    abbrev="bar",
    doc=r"""\
    A bar is 10\ :sup:`5` pascals, around sea level atmospheric pressure.
    """
)
psi = make_compound_unit(
    scale="6.894757e3",
    exponents={pounds: 1, inches: -2},
    name="psi",
    abbrev="psi",
    doc="""\
    Pounds per square inch are the typical units of pressure in |ucs|, despite 
    not using the base unit of length.
    """
)

__all__ = ["Pressure", "pascals", "bars", "psi"] + list(generated.keys())



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.force import pounds
from pyunitx.length import Length, meters, feet
from pyunitx.mass import Mass, kilograms
from pyunitx.time import seconds, Time



Energy = make_compound_dimension({Mass: 1, Length: 2, Time: -2}, "Energy")

joules = make_compound_unit(
    name="joules",
    scale=1,
    exponents={kilograms: 1, meters: 2, seconds: -2},
    abbrev="J",
    doc="""\
    The joule is the base unit of energy in the SI system. It is the amount of
    work done on an object by pushing it with a 1 N force for 1 m. 
    """
)

generated = si_unit(base_unit=joules)
globals().update(generated)

calorie = make_compound_unit(
    name="calorie",
    scale="4.184",
    exponents=joules.composition.to_pairs(),
    abbrev="cal",
    doc="""\
    This is the gram calorie, the amount of heat energy necessary to heat 1 gram
    of water by 1 \xb0C. Nutritional calories are 1000 times larger, using a 
    kilogram in the calculation instead.
    """
)
btu = make_compound_unit(
    name="btu",
    abbrev="btu",
    scale="1.05435e3",
    exponents={feet: 1, pounds: 1},
    doc="""\
    There are many different possible definitions of the British Thermal Unit, 
    this is the thermochemical definition.
    """
)

__all__ = [
              "Energy",
              "joules",
              "calorie",
              "btu",
          ] + list(generated.keys())



"""Conversions between these units are deltas, as the systems have different zeros."""
from decimal import Decimal

from pyunitx._api import make_dimension, make_unit, si_unit

Temperature = make_dimension("Temperature")

kelvin = make_unit(
    name="kelvin",
    dimension=Temperature,
    scale=1,
    abbrev="K",
    doc="""\
    Within this module, all temperatures are treated as deltas, ignoring the
    fact that the systems have different zeros. Thus, calling 
    ``celsius(25).to_kelvin()`` will *not* produce ``kelvin(298.15)`` as you
    might expect, but rather ``kelvin(25)``. To help out, the value of absolute
    zero has been provided in each of the non-absolute units.
    
    Kelvin is the absolute measure of temperature in the SI system.
    Its units are equal in size to the Celsius degree, but its zero point 
    is at absolute zero.
    """
)

generated = si_unit(base_unit=kelvin)
globals().update(generated)

celsius = make_unit(
    name="celsius",
    dimension=Temperature,
    scale=1,
    abbrev="\xb0C",
    doc="""\
    Within this module, all temperatures are treated as deltas, ignoring the
    fact that the systems have different zeros. Thus, calling 
    ``celsius(25).to_kelvin()`` will *not* produce ``kelvin(298.15)`` as you
    might expect, but rather ``kelvin(25)``. To help out, the value of absolute
    zero has been provided in each of the non-absolute units.
    
    To convert from a direct measurement in degrees Celsius to one in Kelvin,
    use ``(<celsius measurement> - absolute_zero_celsius).to_kelvin()``.
    The ``to_kelvin`` at the end is of course not strictly necessary, depending
    on how you are interpreting the results, but is certainly in line with 
    intention.
    """
)
fahrenheit = make_unit(
    name="fahrenheit",
    dimension=Temperature,
    scale=Decimal(5) / 9,
    abbrev="\xb0F",
    doc="""\
    Within this module, all temperatures are treated as deltas, ignoring the
    fact that the systems have different zeros. Thus, calling 
    ``fahrenheit(25).to_rankine()`` will *not* produce ``rankine(484.67)`` as 
    you might expect, but rather ``rankine(25)``. To help out, the value of 
    absolute zero has been provided in each of the non-absolute units.
    
    To convert from a direct measurement in degrees Fahrenheit to one in degrees 
    Rankine, use 
    ``(<fahrenheit measurement> - absolute_zero_fahrenheit).to_rankine()``.
    The ``to_rankine`` at the end is of course not strictly necessary, depending
    on how you are interpreting the results, but is certainly in line with 
    intention.
    """
)
rankine = make_unit(
    name="rankine",
    dimension=Temperature,
    scale=Decimal(5) / 9,
    abbrev="\xb0R",
    doc="""\
    Within this module, all temperatures are treated as deltas, ignoring the
    fact that the systems have different zeros. Thus, calling 
    ``fahrenheit(25).to_rankine()`` will *not* produce ``rankine(484.67)`` as 
    you might expect, but rather ``rankine(25)``. To help out, the value of 
    absolute zero has been provided in each of the non-absolute units.
    
    Degrees Rankine are the US Customary scale's absolute temperature scale.
    Its degrees are the same size as Fahrenheit but its zero is at absolute 
    zero.
    """
)

absolute_zero_celsius = celsius("-273.15")
absolute_zero_fahrenheit = fahrenheit("-459.67")


def celsius_to_kelvin_absolute(degc: celsius) -> kelvin:
    return (degc - absolute_zero_celsius).to_kelvin()


__all__ = [
              "Temperature",
              "kelvin",
              "celsius",
              "fahrenheit",
              "rankine",
              "absolute_zero_celsius",
              "celsius_to_kelvin_absolute",
          ] + list(generated.keys())



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.energy import Energy, joules
from pyunitx.force import pounds
from pyunitx.length import feet
from pyunitx.time import Time, seconds



Power = make_compound_dimension({Energy: 1, Time: -1})

watts = make_compound_unit(
    name="watts",
    scale=1,
    exponents={joules: 1, seconds: -1},
    abbrev="W",
    doc="""\
    A watt is a joule per second. In most cases, like calculating the energy
    expenditure of a heating system, collecting rates of energy use is the most
    useful then multiplying by time to get overall energy usage.
    """
)

generated = si_unit(base_unit=watts)
globals().update(generated)

kilowatts = make_compound_unit(
    name="kilowatts",
    scale=1000,
    exponents={joules: 1, seconds: -1},
    abbrev="kW",
    doc="""\
    A kilowatt is 1000 joules per second. Most household appliances draw power
    on the order of kilowatts.
    """
)

horsepower = make_compound_unit(
    name="horsepower",
    scale="7.456999e2",
    exponents={feet: 1, pounds: 1, seconds: -1},
    abbrev="hp",
    doc="""
    One horsepower is 550 foot-pounds per second, and is produced by a very 
    standards-compliant equine.
    """
)
__all__ = [
              "Power",
              "watts",
              "kilowatts",
              "horsepower",
          ] + list(generated.keys())



import math
from decimal import Decimal

from pyunitx._api import make_dimension, make_unit

__all__ = [
    "Angle",
    "degree",
    "arcminute",
    "arcsecond",
]

Angle = make_dimension("Angle")

# Radians are better to be defined as a base unit but are dimensionless
degree = make_unit(
    name="degree",
    dimension=Angle,
    scale=1,
    abbrev="\xb0",
    doc="""\
    Degrees are a measure of angles. 360 degrees extend around a circle. The 
    true base unit of angle is the dimensionless radian, but degrees see common 
    use as well.
    
    All angle measurements in this module are equipped with a to_radians 
    method alongside the other conversion functions that work between the angle
    measurements defined here. 
    """
)
arcminute = make_unit(
    name="arcminute",
    dimension=Angle,
    scale=Decimal(1) / 60,
    abbrev="\u2032",
    doc="""An arcminute is 1/60 of a degree."""
)
arcsecond = make_unit(
    name="arcsecond",
    dimension=Angle,
    scale=Decimal(1) / 3600,
    abbrev="\u2033",
    doc="""An arcsecond is 1/60 of an arcminute, and therefore 1/3600 of a degree."""
)


def __to_rad(deg):
    return deg.value * deg.scale * Decimal(math.pi) / 180


Angle.to_radians = __to_rad



"""A module for the first-class manipulation of physical quantities.

It is named ``pyunitx`` as a reference to the ``siunitx`` LaTeX package.

Most conversion factors come from
`this document from NIST <https://physics.nist.gov/cuu/pdf/sp811.pdf>`_.
Most values of constants come from
`this list also from NIST <https://physics.nist.gov/cuu/Constants/Table/allascii.txt>`_.
"""

# After the update for automatic conversion functions, this is somewhat less
# important, at least in terms of "not knowing about the conversion functions".
# However, it is still nice to have all the pre-constructed units and dimensions
# with special names available.
# noinspection PyUnresolvedReferences
from . import (
    angle,
    area,
    charge,
    constants,
    current,
    data,
    derived,
    energy,
    force,
    frequency,
    length,
    mass,
    mole,
    power,
    pressure,
    resistance,
    temperature,
    time,
    volume,
    voltage,
)

from ._api import (
    make_dimension,
    make_unit,
    make_compound_dimension,
    make_compound_unit,
    SIUNITX_OLD,
    SIUNITX_NEW,
)



from pyunitx._api import make_compound_dimension, make_compound_unit
from pyunitx.length import Length, meters

__all__ = [
    "Area",
    "meters_squared",
    "acres",
    "hectares",
]

Area = make_compound_dimension(name="Area", exponents={Length: 2})

# no need to explain square meters
meters_squared = make_compound_unit(
    exponents={meters: 2},
    scale=1
)

acres = make_compound_unit(
    name="acres",
    abbrev="acre",
    scale="4.046873e3",
    exponents={meters: 2},
    doc="""\
    An acre is a common measure of land area.
    """
)
hectares = make_compound_unit(
    name="hectares",
    abbrev="ha",
    scale="1e4",
    exponents={meters: 2},
    doc="""\
    A hectare is a unit of land area compatible with SI units, being defined as
    10,000 square meters or the area of a square 100 meters on a side.
    """
)



from pyunitx._api import make_dimension, make_unit, si_unit

Quantity = make_dimension("Quantity")

moles = make_unit(
    name="moles",
    dimension=Quantity,
    scale=1,
    abbrev="mol",
    doc="""\
    A mole is the number of atoms in 12 grams of carbon-12.
    
    It basically is a dimensionless quantity, but is included in the SI system
    for its pragmatic usefulness in chemistry and fluid physics.
    """
)
generated = si_unit(base_unit=moles)
globals().update(generated)

__all__ = [
              "Quantity",
              "moles",
          ] + list(generated.keys())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyunitx']

package_data = \
{'': ['*']}

install_requires = \
['sigfig>=1.3.2,<2.0.0']

extras_require = \
{'docs': ['pypandoc>=1.9,<2.0']}

entry_points = \
{'console_scripts': ['uconvert = pyunitx.uconvert:main']}

setup_kwargs = {
    'name': 'pyunitx',
    'version': '0.6.3',
    'description': 'First-class manipulation of physical quantities',
    'long_description': '# pyunitx\n\n[![Coverage Status](https://coveralls.io/repos/github/the-nick-of-time/units/badge.svg?branch=main)](https://coveralls.io/github/the-nick-of-time/units?branch=main)\n[![Documentation Status](https://readthedocs.org/projects/pyunitx/badge/?version=latest)](https://pyunitx.readthedocs.io/en/latest/?badge=latest)\n\nWhen doing calculations using physical measurements, it\'s all too easy to forget to account for\nunits. This can result in problems when you find you\'ve been adding kilograms to newtons and\nyour calculation is off by a factor of ten.\n\nThis library uses the standard\nlibrary [decimal.Decimal](https://docs.python.org/3/library/decimal.html) for all calculations\nto avoid most floating-point calculation pitfalls. Values given for units are automatically\nconverted so you can enter any value that constructor can take. Functionally, this means that\nfloat notation should be given as strings rather than float literals.\n\nQ. How many meters does light travel in a millisecond?\n\n```pycon\n>>> from pyunitx.time import seconds\n>>> from pyunitx.constants import c\n>>> (c * seconds("1e-3")).sig_figs(5)\n2.9979E+5 m\n\n```\n\nQ. What is that in feet?\n\n```pycon\n>>> from pyunitx.time import seconds\n>>> from pyunitx.constants import c\n>>> (c * seconds("1e-3")).to_feet().sig_figs(5)\n9.8357E+5 ft\n\n```\n\nQ. How fast is someone on the equator moving around the center of the earth?\n\n```pycon\n>>> from pyunitx.time import days\n>>> from pyunitx.constants import earth_radius\n>>> from math import pi\n>>> circumference = 2 * pi * earth_radius\n>>> (circumference / days(1)).to_meters_per_second().sig_figs(3)\n464 m s^-1\n\n```\n\nQ. How fast is the earth orbiting the sun?\n\n```pycon\n>>> from pyunitx.time import julian_years\n>>> from pyunitx.length import au\n>>> from math import pi\n>>> circumference = 2 * pi * au(1)\n>>> (circumference / julian_years(1)).to_kilometers_per_hour().sig_figs(3)\n1.07E+5 km hr^-1\n\n```\n\nQ. What\'s the mass of air in one of your car tires, if the inner radius is 6 inches, the outer\nradius is 12.5 inches, the width is 8 inches, and it\'s filled to 42 psi?[^1]\n\n[^1]: No, I don\'t write homework problems. Why do you ask?\n\n```pycon\n>>> from pyunitx.length import inches\n>>> from pyunitx.pressure import psi\n>>> from pyunitx.constants import R, air_molar_mass\n>>> from pyunitx.temperature import celsius, celsius_to_kelvin_absolute\n>>> from math import pi\n>>> volume = (pi * inches(8) * (inches("12.5") ** 2 - inches(6) ** 2)).to_meters_cubed()\n>>> pressure = psi(42).to_pascals()\n>>> temperature = celsius_to_kelvin_absolute(celsius(25))\n>>> mols = pressure * volume / (R * temperature)\n>>> mass = mols * air_molar_mass\n>>> mass.to_avoirdupois_pounds_mass().sig_figs(3)\n0.369 lbm_A\n\n```\n\nAll constants like `R` are defined in SI base units so you will need to convert your units, but\nas you can see, that task is easy. It\'s just a matter of calling `.to_<other unit>()`. You can\nconvert from any unit to another that measures the same dimension this way. If you\'re going to a\ncomposite unit that hasn\'t been explicitly declared with a name, this is still possible, and the\nlibrary will create a converter for you - you just need to get the name right. The name format\nis as intuitive as possible, as you can see with the above examples.\n\nA name is made of the names of the base units, suffixed with `_squared`, `_cubed`, etc. to\nrelate the size of the exponent and prefixed by `per_` if the exponent is negative. Units with\nnegative exponents are made singular[^2] to follow how you would say it.\n\n[^2]: Naively; it\'s done by just stripping off a trailing \'s\' if there is one.\n\nSome examples of the most complicated possible situations will be illustrative.\n\n```pycon\n>>> from pyunitx.constants import gas_constant, stefan_boltzmann\n>>> print(gas_constant.to_feet_pounds_per_mole_per_rankine().sig_figs(4))\n3.407 ft^2 slug mol^-1 °R^-1 s^-2\n\n>>> print(stefan_boltzmann.to_horsepower_per_feet_squared_per_rankine_to_the_fourth().sig_figs(5))\n3.7013E-10 slug s^-3 °R^-4\n\n```\n\nYou will notice that the output will have all units broken down to their bases. It is guaranteed\nto be equivalent.\n\nNow what happens if a calculation results in a predefined unit, like how newtons times meters\nequals joules?\n\n```pycon\n>>> from pyunitx.voltage import volts\n>>> from pyunitx.resistance import ohms\n>>> print(volts(2) / ohms(100))\n0.02 A\n\n```\n\nCalculations check their result against all the units that have been specially defined to find a\nmatch. However, if you end up with a result that could be broken into some product of complex\nunits (like newton-seconds) this library will *not* do that for you and instead display it in\nits basest components. This is because the number of possible options is large and it\'s not\npossible to figure out what you want.\n\nThis library predefines all the SI units and dimensions, but what if that\'s not enough? You\nmight want to model some other quantity, like cash flow in your budget.\n\n```pycon\n>>> from pyunitx import make_dimension, make_unit\n>>> from pyunitx.time import days\n>>> Money = make_dimension(\'Money\')\n>>> dollars = make_unit(name="dollars", abbrev="$", dimension=Money, scale=1)\n>>> euros = make_unit(name="euros", abbrev="€", dimension=Money, scale="0.98019")\n>>> (dollars(150) / days(7)).to_euros_per_year().sig_figs(6)\n7984.80 € yr^-1\n\n```\n\nFor more examples, including derived units, see the definitions in the package, like\n[energy](https://github.com/the-nick-of-time/units/blob/main/pyunitx/energy.py) or\n[time](https://github.com/the-nick-of-time/units/blob/main/pyunitx/time.py).\n\nThe full documentation can be found at [ReadTheDocs](https://pyunitx.readthedocs.io/en/latest/).\n',
    'author': 'Nick Thurmes',
    'author_email': 'nthurmes@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pyunitx._api import make_compound_dimension, make_compound_unit, si_unit
from pyunitx.length import Length, meters, feet

Volume = make_compound_dimension({Length: 3}, "Volume")

meters_cubed = make_compound_unit(
    scale=1,
    exponents={meters: 3}
)
liters = make_compound_unit(
    name="liters",
    scale="1e-3",
    exponents={meters: 3},
    abbrev="L",
    doc="""\
    A liter is one cubic decimeter, or a cube about as big as the palm of your 
    hand.
    """
)
milliliters = make_compound_unit(
    name="milliliters",
    scale="1e-6",
    exponents={meters: 3},
    abbrev="mL",
    doc="""A milliliter is around 20 drops of water, or the size of a 1-cm cube."""
)

generated = si_unit(base_unit=liters, skip=["milli"])
globals().update(generated)

feet_cubed = make_compound_unit(
    scale=".02831685",
    exponents={feet: 3}
)
fluid_ounces = make_compound_unit(
    name="fluid_ounces",
    exponents={meters: 3},  # As fl oz aren't defined in terms of some particular length cube
    scale="2.957353e-5",
    abbrev="fl oz",
    doc="""\
    Fluid ounces are the closest the |ucs| has to a base unit of volume.
    """
)
fluid_ounce_imperial = make_compound_unit(
    name="imperial fluid ounce",
    exponents={meters: 3},
    scale="2.841306e-5",
)
cups = make_compound_unit(
    name="cups",
    exponents={meters: 3},
    scale="2.365882e-4",
    abbrev="c",
    doc="""\
    Cups are a common unit of measure when cooking. One cup is 8 
    :class:`fluid ounces <fluid_ounces>` or 16 :class:`tablespoons`.
    """
)
teaspoons = make_compound_unit(
    name="teaspoons",
    exponents={meters: 3},
    scale="4.928922e-6",
    abbrev="tsp",
    doc="""\
    Teaspoons are a common unit of measure when cooking. It's about equal to
    5 :class:`milliliters`, if you need to do quick conversion.
    """
)
tablespoons = make_compound_unit(
    name="tablespoons",
    exponents={meters: 3},
    scale="1.478676e-5",
    abbrev="tbsp",
    doc="""\
    Tablespoons are a common unit of measure when cooking.
    One tablespoon is 3 :class:`teaspoons`.
    """
)
gallons = make_compound_unit(
    name="gallons",
    exponents={meters: 3},
    scale="3.785412e-3",
    abbrev="gal",
    doc="""\
    Gallons are the bulk unit of volume in the |ucs|. It is equal to 128
    :class:`fluid ounces <fluid_ounces>`, or 4 quarts, or 8 pints, or 16 
    :class:`cups`. At least this range of measurements is all powers of 2.
    """
)
imperial_gallons = make_compound_unit(
    name="imperial_gallons",
    exponents={meters: 3},
    scale="4.54609e-3"
)

__all__ = [
              "meters_cubed",
              "liters",
              "milliliters",
              "feet_cubed",
              "fluid_ounces",
              "cups",
              "teaspoons",
              "tablespoons",
              "gallons",
          ] + list(generated.keys())



class UnitException(Exception):
    pass


class OperationError(UnitException, TypeError):
    def __init__(self, operation, left, right):
        self.operation = operation
        self.left = left
        self.right = right

    def __repr__(self):
        return f"Cannot {self.operation} {self.left} and {self.right}"


class ImplicitConversionError(UnitException, TypeError):
    def __init__(self, source, dest):
        self.source = source
        self.dest = dest

    def __repr__(self):
        return f"Will not implicitly convert {self.source} to {self.dest}"
